﻿//--------------------------------------------------------------------------------------
//			** Z-fighting Aware Depth Peeling **
//  
//  Descpiption: 
//		(1) Store the fragments placed at the extracted depth layer using a K-buffer.
//
//  Authors: Andreas A. Vasilakis - Ioannis Fudos
//  Emails : abasilak@cs.uoi.gr - fudos@cs.uoi.gr
// 
//  Software is distributed under the following BSD-style license:
//
//  Copyright © 2010, 2011 University of Ioannina. All Rights Reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//
//  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
//  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
//  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//  The work was partially supported by a Heraclitus II grant through the operational programme “Education and Lifelong Learning” 
//  which is co-financed by Greece and the European Union through the European Social Fund.
//--------------------------------------------------------------------------------------

#version 330 core

#if multisample
#extension GL_ARB_sample_shading : enable
#endif

#if packing
#extension GL_ARB_gpu_shader5 : enable
#endif

#define bufferSize 8

vec4 computePixelColor();

#if multisample
	uniform sampler2DMS	tex_depth;
#if packing	
	uniform usampler2DMS tex_peel_0;
	uniform usampler2DMS tex_peel_1;
	uniform usampler2DMS tex_peel_2;
	uniform usampler2DMS tex_peel_3;
	uniform usampler2DMS tex_peel_4;
	uniform usampler2DMS tex_peel_5;
	uniform usampler2DMS tex_peel_6;
	uniform usampler2DMS tex_peel_7;
#else
	uniform sampler2DMS	tex_peel_0;
	uniform sampler2DMS tex_peel_1;
	uniform sampler2DMS tex_peel_2;
	uniform sampler2DMS tex_peel_3;
	uniform sampler2DMS tex_peel_4;
	uniform sampler2DMS tex_peel_5;
	uniform sampler2DMS tex_peel_6;
	uniform sampler2DMS tex_peel_7;
#endif

#else
	uniform sampler2DRect tex_depth;
#if packing
	uniform usampler2DRect tex_peel_0;
	uniform usampler2DRect tex_peel_1;
	uniform usampler2DRect tex_peel_2;
	uniform usampler2DRect tex_peel_3;
	uniform usampler2DRect tex_peel_4;
	uniform usampler2DRect tex_peel_5;
	uniform usampler2DRect tex_peel_6;
	uniform usampler2DRect tex_peel_7;
#else
	uniform sampler2DRect tex_peel_0;
	uniform sampler2DRect tex_peel_1;
	uniform sampler2DRect tex_peel_2;
	uniform sampler2DRect tex_peel_3;
	uniform sampler2DRect tex_peel_4;
	uniform sampler2DRect tex_peel_5;
	uniform sampler2DRect tex_peel_6;
	uniform sampler2DRect tex_peel_7;
#endif

#endif

#if packing
	layout(location = 0, index = 0) out uvec4 out_frag_color[8];
#else
	layout(location = 0, index = 0) out  vec4 out_frag_color[8];
#endif

void main(void)
{
	float depth;
#if multisample
	depth = texelFetch(tex_depth, ivec2(gl_FragCoord.xy), gl_SampleID).r;
#else
	depth = texture(tex_depth, gl_FragCoord.xy).r;
#endif
	if (gl_FragCoord.z != depth)
		discard;
#if packing
		uvec4 k[8];
#else
		vec4  k[8];
#endif

#if multisample
	k[0] = texelFetch(tex_peel_0, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[1] = texelFetch(tex_peel_1, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[2] = texelFetch(tex_peel_2, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[3] = texelFetch(tex_peel_3, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[4] = texelFetch(tex_peel_4, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[5] = texelFetch(tex_peel_5, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[6] = texelFetch(tex_peel_6, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[7] = texelFetch(tex_peel_7, ivec2(gl_FragCoord.xy), gl_SampleID);
#else
	k[0] = texture(tex_peel_0, gl_FragCoord.xy);
	k[1] = texture(tex_peel_1, gl_FragCoord.xy);
	k[2] = texture(tex_peel_2, gl_FragCoord.xy);
	k[3] = texture(tex_peel_3, gl_FragCoord.xy);
	k[4] = texture(tex_peel_4, gl_FragCoord.xy);
	k[5] = texture(tex_peel_5, gl_FragCoord.xy);
	k[6] = texture(tex_peel_6, gl_FragCoord.xy);
	k[7] = texture(tex_peel_7, gl_FragCoord.xy); 
#endif

	int i;

#if packing
	int j;
	bool b = false;
	for(i=0; i<bufferSize; i++){
		for(j=0; j<4; j++)
			if(k[i][j] == 0U)
			{
				k[i][j] = packUnorm4x8(computePixelColor());
				b = true;
				break;
			}
		if(b)
			break;
	}
#else
	for(i=0; i<bufferSize; i++)
		if(k[i].w == 0.0f)
		{
			k[i] = computePixelColor();
			break;
		}
#endif

	for(i=0; i<bufferSize; i++)
		out_frag_color[i] = k[i];
}