﻿//--------------------------------------------------------------------------------------
//			** Z-fighting Aware Depth Peeling **
//  
//  Descpiption: 
//		(1) Extract the color information or (2) the transparent result of the peeled fragments currrently stored at the K-buffer.
//
//  Authors: Andreas A. Vasilakis - Ioannis Fudos
//  Emails : abasilak@cs.uoi.gr - fudos@cs.uoi.gr
// 
//  Software is distributed under the following BSD-style license:
//
//  Copyright © 2010, 2011 University of Ioannina. All Rights Reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//
//  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
//  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
//  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//  The work was partially supported by a Heraclitus II grant through the operational programme “Education and Lifelong Learning” 
//  which is co-financed by Greece and the European Union through the European Social Fund.
//--------------------------------------------------------------------------------------

#version 330 core

#if multisample
#extension GL_ARB_sample_shading : enable
#endif

#if packing
#extension GL_ARB_gpu_shader5 : enable
#endif

#define bufferSize 8

uniform int    layer;
uniform bool   useTransparency;

#if multisample

#if packing
	uniform usampler2DMS tex_peel_0;
	uniform usampler2DMS tex_peel_1;
	uniform usampler2DMS tex_peel_2;
	uniform usampler2DMS tex_peel_3;
	uniform usampler2DMS tex_peel_4;
	uniform usampler2DMS tex_peel_5;
	uniform usampler2DMS tex_peel_6;
	uniform usampler2DMS tex_peel_7;
#else
	uniform sampler2DMS	tex_peel_0;
	uniform sampler2DMS tex_peel_1;
	uniform sampler2DMS tex_peel_2;
	uniform sampler2DMS tex_peel_3;
	uniform sampler2DMS tex_peel_4;
	uniform sampler2DMS tex_peel_5;
	uniform sampler2DMS tex_peel_6;
	uniform sampler2DMS tex_peel_7;
#endif

#else

#if packing
	uniform usampler2DRect tex_peel_0;
	uniform usampler2DRect tex_peel_1;
	uniform usampler2DRect tex_peel_2;
	uniform usampler2DRect tex_peel_3;
	uniform usampler2DRect tex_peel_4;
	uniform usampler2DRect tex_peel_5;
	uniform usampler2DRect tex_peel_6;
	uniform usampler2DRect tex_peel_7;
#else
	uniform sampler2DRect tex_peel_0;
	uniform sampler2DRect tex_peel_1;
	uniform sampler2DRect tex_peel_2;
	uniform sampler2DRect tex_peel_3;
	uniform sampler2DRect tex_peel_4;
	uniform sampler2DRect tex_peel_5;
	uniform sampler2DRect tex_peel_6;
	uniform sampler2DRect tex_peel_7;
#endif

#endif

// OUT
#if packing
	layout(location = 0, index = 0) out uvec4 out_frag_color;
#else
	layout(location = 0, index = 0) out  vec4 out_frag_color;
#endif

#if packing
	uint  resolveAlphaBlend();
	uvec4 k[8];
#else
	vec4  resolveAlphaBlend();
	vec4  k[8];
#endif

void main(void)
{
#if multisample
	k[0] = texelFetch(tex_peel_0, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[1] = texelFetch(tex_peel_1, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[2] = texelFetch(tex_peel_2, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[3] = texelFetch(tex_peel_3, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[4] = texelFetch(tex_peel_4, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[5] = texelFetch(tex_peel_5, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[6] = texelFetch(tex_peel_6, ivec2(gl_FragCoord.xy), gl_SampleID);
	k[7] = texelFetch(tex_peel_7, ivec2(gl_FragCoord.xy), gl_SampleID);
#else
	k[0] = texture(tex_peel_0, gl_FragCoord.xy);
	k[1] = texture(tex_peel_1, gl_FragCoord.xy);
	k[2] = texture(tex_peel_2, gl_FragCoord.xy);
	k[3] = texture(tex_peel_3, gl_FragCoord.xy);
	k[4] = texture(tex_peel_4, gl_FragCoord.xy);
	k[5] = texture(tex_peel_5, gl_FragCoord.xy);
	k[6] = texture(tex_peel_6, gl_FragCoord.xy);
	k[7] = texture(tex_peel_7, gl_FragCoord.xy); 
#endif

#if packing
	out_frag_color.r = (useTransparency) ? resolveAlphaBlend() : k[layer/4][layer%4];
#else
	out_frag_color   = (useTransparency) ? resolveAlphaBlend() : k[layer];
#endif		
}

#if packing
uint resolveAlphaBlend()
{
	vec4 color,finalColor = vec4(0.0f);
	for(int i=0; i<bufferSize; i++)
		for(int j=0; j<4; j++)
		{
			color=unpackUnorm4x8(k[i][j]);
			if(color.a == 0.0f)
				return packUnorm4x8(finalColor);
			finalColor += color*(1.0f-finalColor.a);
		}
	return packUnorm4x8(finalColor);
}
#else
vec4 resolveAlphaBlend()
{
	vec4 finalColor = vec4(0.0f);
	for(int i=0; i<bufferSize; i++)
	{
		if(k[i].a > 0.0f)
			finalColor += k[i]*(1.0f-finalColor.a);
		else
			break;
	}
	return finalColor;
}
#endif